/*
 * Copyright (c) 2007 by Karl Helgason
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package benchmark;

import java.io.BufferedInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Receiver;
import javax.sound.midi.Sequence;
import javax.sound.midi.Soundbank;
import javax.sound.midi.Track;
import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;

import com.sun.media.sound.AudioSynthesizer;
import com.sun.media.sound.SoftSynthesizer;

public class Benchmark {

	public static double send(Sequence seq, Receiver recv) {
		float divtype = seq.getDivisionType();
		assert (seq.getDivisionType() == Sequence.PPQ);
		Track[] tracks = seq.getTracks();
		int[] trackspos = new int[tracks.length];
		int mpq = 60000000 / 100;
		int seqres = seq.getResolution();
		long lasttick = 0;
		long curtime = 0;
		while (true) {
			MidiEvent selevent = null;
			int seltrack = -1;
			for (int i = 0; i < tracks.length; i++) {
				int trackpos = trackspos[i];
				Track track = tracks[i];
				if (trackpos < track.size()) {
					MidiEvent event = track.get(trackpos);
					if (selevent == null
							|| event.getTick() < selevent.getTick()) {
						selevent = event;
						seltrack = i;
					}
				}
			}
			if (seltrack == -1)
				break;
			trackspos[seltrack]++;
			long tick = selevent.getTick();
			if (divtype == Sequence.PPQ)
				curtime += ((tick - lasttick) * mpq) / seqres;
			else
				curtime = (long) ((tick * 1000000.0 * divtype) / seqres);
			lasttick = tick;
			MidiMessage msg = selevent.getMessage();
			if (msg instanceof MetaMessage) {
				if (divtype == Sequence.PPQ)
					if (((MetaMessage) msg).getType() == 0x51) {
						byte[] data = ((MetaMessage) msg).getData();
						mpq = ((data[0] & 0xff) << 16)
								| ((data[1] & 0xff) << 8) | (data[2] & 0xff);
					}
			} else {
				if (recv != null)
					recv.send(msg, curtime);
			}
		}

		return curtime / 1000000.0;
	}

	Soundbank soundbank;

	Sequence sequence;

	double total;

	public Benchmark() throws Exception {
		InputStream sb = new BufferedInputStream(Benchmark.class
				.getResourceAsStream("/benchmark/ding.sf2"));
		soundbank = MidiSystem.getSoundbank(sb);
		sb.close();

		InputStream si = new BufferedInputStream(Benchmark.class
				.getResourceAsStream("/benchmark/expresso.mid"));
		sequence = MidiSystem.getSequence(si);
		si.close();

		total = send(sequence, null);
	}

	public double totalTime() {
		return total;
	}

	public double benchmark(AudioFormat format, Map<String, Object> info)
			throws Exception {
		OutputStream nullout = new OutputStream() {
			public void write(int b) throws IOException {
			}

			public void write(byte[] b, int off, int len) throws IOException {
			}

			public void write(byte[] b) throws IOException {
			}
		};
		return render(nullout, format, info);
	}

	public double render(OutputStream os, AudioFormat format,
			Map<String, Object> info) throws Exception {
		AudioSynthesizer synth = (AudioSynthesizer) new SoftSynthesizer();
		AudioInputStream stream = synth.openStream(format, info);
		Receiver recv = synth.getReceiver();
		Soundbank defsbk = synth.getDefaultSoundbank();
		if (defsbk != null)
			synth.unloadAllInstruments(defsbk);
		synth.loadAllInstruments(soundbank);

		send(sequence, recv);

		long len = (long) (stream.getFormat().getFrameRate() * (totalTime() + 4));
		stream = new AudioInputStream(stream, stream.getFormat(), len);

		long t = System.currentTimeMillis();
		AudioSystem.write(stream, AudioFileFormat.Type.WAVE, os);
		t = System.currentTimeMillis() - t;
		stream.close();
		return (t / 1000.0);
	}

	public static void main(String[] args) throws Exception {
		if (args.length != 0) {
			Benchmark benchmark = new Benchmark();
			FileOutputStream fos = new FileOutputStream(args[0]);
			benchmark.render(fos, null, null);
			fos.close();
			System.exit(0);
			return;
		}

		System.out.println(System.getProperty("java.runtime.name") + ", "
				+ System.getProperty("java.runtime.version"));
		System.out.println(System.getProperty("os.arch") + ", "
				+ System.getProperty("os.name") + ", "
				+ System.getProperty("os.version"));
		System.out.println("Total Memory: "
				+ Runtime.getRuntime().totalMemory() / 1048576 + "M");
		System.out.println("Free Memory: " + Runtime.getRuntime().freeMemory()
				/ 1048576 + "M");
		System.out.println("Max Memory: " + Runtime.getRuntime().maxMemory()
				/ 1048576 + "M");
		System.out.println("Available Processors: "
				+ Runtime.getRuntime().availableProcessors());
		System.out.println();

		AudioFormat format = new AudioFormat(44100, 16, 2, true, false);
		Map<String, Object> info = new HashMap<String, Object>();
		// info.put("control rate", 40f);
		// info.put("interpolation", "sinc");

		DecimalFormat dformat = new DecimalFormat("###0.00");
		dformat.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));

		Benchmark benchmark = new Benchmark();
		double totallen = benchmark.totalTime();
		double t = benchmark.benchmark(format, info);
		System.out.println("SCORE: " + dformat.format(totallen / t));

		double defaulttime = t;

		System.out.println();
		{
			format = new AudioFormat(11025, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("11025 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(22050, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("22050 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(44100, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(88200, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("88200 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
		}
		System.out.println();

		{
			format = new AudioFormat(44100, 8, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz  8 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(44100, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(44100, 24, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 24 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(44100, 32, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 32 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
		}
		System.out.println();

		{
			format = new AudioFormat(44100, 16, 1, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 16 bit 1 ch: x"
					+ dformat.format(defaulttime / t));
			format = new AudioFormat(44100, 16, 2, true, false);
			t = benchmark.benchmark(format, info);
			System.out.println("44100 Hz 16 bit 2 ch: x"
					+ dformat.format(defaulttime / t));
		}

		System.out.println();
		{
			format = new AudioFormat(44100, 16, 2, true, false);

			info.clear();
			info.put("control rate", 50f);
			t = benchmark.benchmark(format, info);
			System.out.println("control rate = 50 Hz: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("control rate", 100f);
			t = benchmark.benchmark(format, info);
			System.out.println("control rate = 100 Hz: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("control rate", 147f);
			t = benchmark.benchmark(format, info);
			System.out.println("control rate = 147 Hz: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("control rate", 200f);
			t = benchmark.benchmark(format, info);
			System.out.println("control rate = 200 Hz: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("control rate", 250f);
			t = benchmark.benchmark(format, info);
			System.out.println("control rate = 250 Hz: x"
					+ dformat.format(defaulttime / t));

		}

		System.out.println();
		{
			format = new AudioFormat(44100, 16, 2, true, false);

			info.clear();
			info.put("interpolation", "point");
			t = benchmark.benchmark(format, info);
			System.out.println("interpolation = point: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("interpolation", "linear");
			t = benchmark.benchmark(format, info);
			System.out.println("interpolation = linear: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("interpolation", "cubic");
			t = benchmark.benchmark(format, info);
			System.out.println("interpolation = cubic: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("interpolation", "lanczos");
			t = benchmark.benchmark(format, info);
			System.out.println("interpolation = lanczos: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("interpolation", "sinc");
			t = benchmark.benchmark(format, info);
			System.out.println("interpolation = sinc: x"
					+ dformat.format(defaulttime / t));

		}

		System.out.println();
		{
			format = new AudioFormat(44100, 16, 2, true, false);

			info.clear();
			info.put("max polyphony", 4);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 4: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("max polyphony", 8);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 8: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("max polyphony", 16);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 16: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("max polyphony", 32);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 32: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("max polyphony", 64);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 64: x"
					+ dformat.format(defaulttime / t));

			info.clear();
			info.put("max polyphony", 128);
			t = benchmark.benchmark(format, info);
			System.out.println("max polyphony = 128: x"
					+ dformat.format(defaulttime / t));

		}

		System.exit(0);

	}
}
